#!/usr/bin/env python3
import logging
from pathlib import Path

import click
from sqlalchemy import create_engine

from .update_usagi import update_usagi_file

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--url",
    required=True,
    help="SQLAlchemy database URL",
    type=click.STRING,
)
@click.option(
    "--schema",
    required=True,
    help="Schema containing the OHDSI vocabulary tables",
    type=click.STRING,
)
@click.option(
    "-f",
    "--usagi-file",
    required=True,
    help="A file saved or exported from Usagi",
    type=click.Path(dir_okay=False, exists=True, readable=True, path_type=Path),
)
@click.option(
    "-h",
    "--allow-homonyms",
    is_flag=True,
    default=False,
    help="If no standard concept can be found through concept relationships, "
    "try to find it through concepts with the same concept name. It is "
    "strongly recommended to index the concept_name column when using this.",
)
@click.option(
    "-i",
    "--ignore-case",
    is_flag=True,
    default=False,
    help="When searching for homonyms, do so in a case-insensitive way. "
    "It is strongly recommended to create a functional index on the "
    "concept table for this.",
)
@click.option(
    "-m",
    "--write-map-paths",
    is_flag=True,
    default=False,
    help="Write an additional file that shows for each remapped concept, "
    "through which mapping relationships the new target concept was determined.",
)
@click.option(
    "-s",
    "--inspect-only",
    is_flag=True,
    default=False,
    help="Instead of writing updated mappings to a file, only show results in the console.",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite the existing Usagi file (otherwise a new file is written).",
)
@click.option(
    "-a",
    "--update-all",
    is_flag=True,
    default=False,
    help="Also update fields of concepts that don't get a new mapping, but "
    "do have outdated properties (e.g. changed domain_id)",
)
def _update_usagi_cli(
    url: str,
    schema: str,
    usagi_file: Path,
    allow_homonyms: bool,
    ignore_case: bool,
    write_map_paths: bool,
    inspect_only: bool,
    overwrite: bool,
    update_all: bool,
) -> None:
    """
    Parse an Usagi saved/exported file to update non-standard concepts.
    """
    engine = create_engine(url)
    update_usagi_file(
        engine,
        schema,
        usagi_file,
        allow_homonyms,
        ignore_case,
        write_map_paths,
        inspect_only,
        overwrite,
        update_all,
    )


def main():
    _update_usagi_cli()


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest
from omop_cdm.regular.cdm54 import Concept
from sqlalchemy import Engine

from kotobuki.mapping_updater.relationship import MapLink, NewMap, Relationship
from kotobuki.mapping_updater.update_usagi import update_usagi_file
from tests.python.mapping_updater.conftest import (
    USAGI_STCM_FILE,
    read_yaml_file,
    write_tmp_usagi_file,
)

TARGET_CONCEPT1 = Concept(
    concept_id=3,
    concept_name="The gold standard",
    vocabulary_id="V1",
    domain_id="D1",
)
TARGET_CONCEPT2 = Concept(
    concept_id=30,
    concept_name="The other gold standard",
    vocabulary_id="V1",
    domain_id="D1",
)
VALUE_CONCEPT1 = Concept(
    concept_id=40,
    concept_name="The gold value",
    vocabulary_id="V1",
    domain_id="Meas Value",
)
SOURCE_CONCEPT1 = Concept(
    concept_id=2,
    concept_name="The gold non-standard",
    vocabulary_id="V1",
    domain_id="D1",
)
HOMONYM_CONCEPT1 = Concept(
    concept_id=99,
    concept_name="THE GOLD NON-STANDARD",
    vocabulary_id="V2",
    domain_id="D1",
)


def test_simple_mapping_path():
    map_links = [MapLink(concept=SOURCE_CONCEPT1)]
    new_map = NewMap(concepts=[TARGET_CONCEPT1], map_path=map_links)
    assert new_map.to_map_path_data() == {
        "2 The gold non-standard (V1 - D1)": {
            "maps_to": ["3 The gold standard (V1 - D1)"],
        }
    }


def test_via_homonym_mapping_path():
    map_links = [
        MapLink(concept=SOURCE_CONCEPT1),
        MapLink(concept=HOMONYM_CONCEPT1, via=Relationship.HOMONYM),
    ]
    new_map = NewMap(concepts=[TARGET_CONCEPT1], map_path=map_links)
    assert new_map.to_map_path_data() == {
        "2 The gold non-standard (V1 - D1)": {
            "map_path": ["(Homonym) 99 THE GOLD NON-STANDARD"],
            "maps_to": ["3 The gold standard (V1 - D1)"],
        }
    }


def test_one_to_two_mapping_path():
    map_links = [MapLink(concept=SOURCE_CONCEPT1)]
    new_map = NewMap(concepts=[TARGET_CONCEPT1, TARGET_CONCEPT2], map_path=map_links)
    assert new_map.to_map_path_data() == {
        "2 The gold non-standard (V1 - D1)": {
            "maps_to": ["3 The gold standard (V1 - D1)", "30 The other gold standard (V1 - D1)"],
        }
    }


def test_maps_to_value_mapping_path():
    map_links = [MapLink(concept=SOURCE_CONCEPT1)]
    new_map = NewMap(
        concepts=[TARGET_CONCEPT1],
        value_as_concept=[VALUE_CONCEPT1],
        map_path=map_links,
    )
    assert new_map.to_map_path_data() == {
        "2 The gold non-standard (V1 - D1)": {
            "maps_to": ["3 The gold standard (V1 - D1)"],
            "maps_to_value": ["40 The gold value (V1 - Meas Value)"],
        }
    }


@pytest.mark.usefixtures("create_vocab_tables")
def test_write_map_paths_e2e(tmp_path: Path, pg_db_engine: Engine):
    tmp_usagi_file = write_tmp_usagi_file(tmp_path, USAGI_STCM_FILE)
    update_usagi_file(pg_db_engine, "vocab", tmp_usagi_file, write_map_paths=True)
    map_paths_file = next(tmp_path.glob("*.yml"))
    contents = read_yaml_file(map_paths_file)
    expected_contents = {
        "1 x deprecated1 (0 - 0)": {
            "map_path": ["(Concept replaced by) 2 x deprecated2"],
            "maps_to": ["3 x standard (0 - 0)"],
        },
        "2 x deprecated2 (0 - 0)": {"maps_to": ["3 x standard (0 - 0)"]},
        "7 History of halitosis (0 - 0)": {
            "maps_to": ["8 History of (0 - 0)"],
            "maps_to_value": ["9 Halitosis (0 - 0)"],
        },
        "10 one_to_two_map (0 - 0)": {"maps_to": ["11 1/2 (0 - 0)", "12 2/2 (0 - 0)"]},
        "13 dep_concept_two_values (0 - 0)": {
            "maps_to": ["14 superstandard (0 - 0)"],
            "maps_to_value": ["15 val1 (0 - 0)", "16 val2 (0 - 0)"],
        },
    }
    assert expected_contents == contents



import pytest
from omop_cdm.regular.cdm54 import Concept
from sqlalchemy import Engine, select
from sqlalchemy.orm import Session

from kotobuki.mapping_updater.db import find_new_mapping
from kotobuki.mapping_updater.relationship import NewMap

pytestmark = pytest.mark.usefixtures("create_vocab_tables")


def get_concept_by_id(session: Session, concept_id: int) -> Concept | None:
    return session.scalars(select(Concept).filter(Concept.concept_id == concept_id)).one_or_none()


def get_new_map(
    concept_id: int,
    engine: Engine,
    homonyms: bool = False,
    ignore_case: bool = False,
) -> NewMap | None:
    with Session(engine, expire_on_commit=False) as session, session.begin():
        concept = get_concept_by_id(session, concept_id=concept_id)
        return find_new_mapping(
            concept=concept,
            search_homonyms=homonyms,
            ignore_case=ignore_case,
            session=session,
        )


def test_no_map(pg_db_engine: Engine):
    """Concept with no relationships cannot be updated."""
    result = get_new_map(concept_id=4, engine=pg_db_engine)
    assert result is None


def test_simple_maps_to_standard(pg_db_engine: Engine):
    """Concept 2 maps to concept 3 directly (Maps to)."""
    result = get_new_map(concept_id=2, engine=pg_db_engine)
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 3
    assert not result.value_as_concept


def test_two_step_mapping(pg_db_engine: Engine):
    """Concept 1 is replaced by 2 and then maps to 3."""
    result = get_new_map(concept_id=1, engine=pg_db_engine)
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 3
    assert not result.value_as_concept


def test_homonym_mapping_inactive(pg_db_engine: Engine):
    """Concept 5 can only map to 6 via a homonym."""
    result = get_new_map(concept_id=5, engine=pg_db_engine, homonyms=False)
    # Without activating homonym search, no mapping should be found
    assert result is None


def test_maps_to_value(pg_db_engine: Engine):
    """Concept 7 has both maps to and maps to value relationships."""
    result = get_new_map(concept_id=7, engine=pg_db_engine)
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 8
    assert len(result.value_as_concept) == 1
    assert result.value_as_concept[0].concept_id == 9


def test_one_maps_to_two(pg_db_engine: Engine):
    """Concept 10 has maps to relationships to 11 and 12."""
    result = get_new_map(concept_id=10, engine=pg_db_engine)
    assert len(result.concepts) == 2
    assert {c.concept_id for c in result.concepts} == {11, 12}
    assert not result.value_as_concept


def test_multiple_maps_to_value(pg_db_engine: Engine):
    """Concept 13 has maps to value relationships to 15 and 16."""
    result = get_new_map(concept_id=13, engine=pg_db_engine)
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 14
    assert {c.concept_id for c in result.value_as_concept} == {15, 16}



import shutil
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import pytest
import yaml
from omop_cdm.constants import VOCAB_SCHEMA
from sqlalchemy import Connection, Engine, create_engine, text
from sqlalchemy.sql.ddl import CreateSchema, DropSchema
from testcontainers.postgres import PostgresContainer

POSTGRES_IMAGE = "postgres:16-alpine"
SCHEMA_MAP: dict[str, str] = {VOCAB_SCHEMA: "vocab"}

DB_SETUP_DIR = Path(__file__).parent / "db_setup"
VOCAB_DDL_FILE = DB_SETUP_DIR / "vocab_ddl.sql"
VOCAB_DATA_DIR = DB_SETUP_DIR / "vocab_data"

TEST_DATA_DIR = Path(__file__).parent / "test_data"
MAP_TO_0_USAGI_FILE = TEST_DATA_DIR / "usagi_files" / "mapping_to_0.csv"
USAGI_STCM_FILE = TEST_DATA_DIR / "stcm" / "usagi_test_stcm.csv"


@pytest.fixture(scope="session")
def pg_db_engine() -> Iterator[Engine]:
    with PostgresContainer(POSTGRES_IMAGE) as postgres:
        engine = create_engine(postgres.get_connection_url())
        engine = engine.execution_options(schema_translate_map=SCHEMA_MAP)
        yield engine


def create_schemas(schemas: set[str], conn: Connection) -> None:
    for schema in schemas:
        conn.execute(CreateSchema(schema, if_not_exists=True))


def drop_schemas(schemas: set[str], conn: Connection) -> None:
    for schema in schemas:
        conn.execute(DropSchema(schema, cascade=True, if_exists=True))


@contextmanager
def temp_schemas(engine: Engine, schemas: set[str]):
    with engine.begin() as conn:
        create_schemas(schemas, conn)
    yield
    with engine.begin() as conn:
        drop_schemas(schemas, conn)


@pytest.fixture(scope="session")
def create_vocab_tables(pg_db_engine: Engine):
    """Create vocabulary tables subset and load test data."""
    with temp_schemas(pg_db_engine, schemas=SCHEMA_MAP.values()):
        create_vocab_subset(pg_db_engine)
        load_relationship_test_data(pg_db_engine)
        yield pg_db_engine


def create_vocab_subset(engine: Engine):
    ddl_queries = text(VOCAB_DDL_FILE.read_text(encoding="utf8"))
    with engine.begin() as con:
        con.execute(ddl_queries)


def load_relationship_test_data(engine: Engine) -> None:
    for vocab_file in VOCAB_DATA_DIR.glob("*.csv"):
        table = f"vocab.{vocab_file.stem}"
        connection = engine.raw_connection()
        try:
            cursor = connection.cursor()
            statement = f"COPY {table} FROM STDIN WITH DELIMITER E'\t' CSV HEADER QUOTE E'\b';"
            with vocab_file.open("rb") as f:
                cursor.copy_expert(sql=statement, file=f)
            cursor.close()
            connection.commit()
        finally:
            connection.close()


def write_tmp_usagi_file(tmp_path: Path, original_usagi_file: Path) -> Path:
    tmp_usagi_file = tmp_path / original_usagi_file.name
    shutil.copyfile(original_usagi_file, tmp_usagi_file)
    return tmp_usagi_file


def read_yaml_file(path: Path) -> dict:
    """Read the YAML file and return as dict."""
    with path.open("rt") as f:
        return yaml.safe_load(f.read())



import csv
import logging
from collections.abc import Collection
from pathlib import Path
from time import strftime, time
from typing import Any

import pandas as pd
import yaml
from omop_cdm.regular.cdm54 import Concept

from .relationship import NewMap

logger = logging.getLogger(__name__)


CONCEPT_ID_COLUMNS = {
    "conceptId",  # Usagi save file
    "targetConceptId",  # Usagi review file
    "target_concept_id",  # STCM file
}

USAGI_DATE_FORMAT = "%Y%m%d"


def _get_concept_col(header: pd.Index) -> str:
    for col in CONCEPT_ID_COLUMNS:
        if col in header:
            return col
    raise ValueError(
        f"Could not find column containing concept_id values. "
        f"Allowed columns: {CONCEPT_ID_COLUMNS}"
    )


def get_target_concepts(usagi_file: Path) -> set[int]:
    """Get set of target concept_ids from an Usagi file."""
    concept_col = _get_concept_col(pd.read_csv(usagi_file, nrows=0).columns)
    df = pd.read_csv(usagi_file, usecols=[concept_col], dtype="Int32", index_col=False)
    df.dropna(inplace=True)
    concept_ids = set(df[concept_col].unique())
    # convert numpy int32 to regular python int
    concept_ids = {n.item() for n in concept_ids}
    # Ignore any zeroes
    concept_ids.discard(0)
    return concept_ids


def to_int(value: Any) -> int | None:
    """Convert value to int if possible, else return None."""
    try:
        return int(value)
    except (ValueError, TypeError):
        return None


def _set_field_values(
    row: dict[str, str], fields: Collection[str], new_value: str
) -> dict[str, str]:
    for field in fields:
        if field in row:
            row[field] = new_value
    return row


def update_row(
    row: dict[str, str],
    target_concept: Concept,
    is_new: bool,
) -> dict[str, str]:
    """Update all Usagi file fields where applicable."""
    # The list below is a superset of the fields in Usagi
    # save/review/STCM files. As these are sometimes manually
    # added/modified, the code below simply checks for all possible
    # fields whether they are present in the original Usagi file and
    # updates values where needed.

    # These fields should only be set when the old mapping is replaced
    if is_new:
        row = _set_field_values(row, CONCEPT_ID_COLUMNS, f"{target_concept.concept_id}")
        row = _set_field_values(row, ["mappingStatus"], "UNCHECKED")
        row = _set_field_values(row, ["equivalence"], "UNREVIEWED")
        row = _set_field_values(row, ["statusSetBy", "createdBy"], "UpdateBot")
        row = _set_field_values(row, ["statusSetOn", "createdOn"], f"{int(time())}")

    fields = ["conceptName", "targetConceptName"]
    row = _set_field_values(row, fields, f"{target_concept.concept_name}")
    row = _set_field_values(row, ["domainId", "targetDomainId"], f"{target_concept.domain_id}")
    fields = ["targetVocabularyId", "target_vocabulary_id"]
    row = _set_field_values(row, fields, f"{target_concept.vocabulary_id}")
    sc = target_concept.standard_concept
    row = _set_field_values(row, ["targetStandardConcept"], f"{'' if sc is None else sc}")
    row = _set_field_values(row, ["targetChildCount", "targetParentCount"], "")
    row = _set_field_values(row, ["targetConceptClassId"], f"{target_concept.concept_class_id}")
    row = _set_field_values(row, ["targetConceptCode"], f"{target_concept.concept_code}")
    start_date = f"{target_concept.valid_start_date.strftime(USAGI_DATE_FORMAT)}"
    row = _set_field_values(row, ["targetValidStartDate"], start_date)
    end_date = f"{target_concept.valid_end_date.strftime(USAGI_DATE_FORMAT)}"
    row = _set_field_values(row, ["targetValidEndDate"], end_date)
    ir = target_concept.invalid_reason
    return _set_field_values(row, ["targetInvalidReason"], f"{'' if ir is None else ir}")


def get_new_lines(
    line: dict[str, str],
    new_mappings: dict[int, NewMap | None],
    concept_lookup: dict[int, Concept] | None = None,
) -> list[dict[str, str]]:
    """Return updated Usagi file line(s)."""
    target_concept_id_col = next(col for col in CONCEPT_ID_COLUMNS if col in line)
    target_concept_id = to_int(line[target_concept_id_col])
    new_map = new_mappings.get(target_concept_id)
    new_line = line.copy()
    new_lines = []

    # Target concept was non-standard; new mappings available
    if new_map is not None:
        for concept in new_map.concepts:
            new_lines.append(update_row(new_line.copy(), concept, is_new=True))

        # If there are also MAPS_TO_VALUE relationships, add more lines
        if new_map.value_as_concept:
            for value_concept in new_map.value_as_concept:
                value_map_line = line.copy()
                value_map_line["mappingType"] = "MAPS_TO_VALUE"
                new_lines.append(update_row(value_map_line, value_concept, is_new=True))
        return new_lines

    # No new mappings; update concept properties if concept_lookup is provided.
    # It can also be that the target_concept_id doesn't exist in the user's
    # database, therefore check if concept exists in concept_lookup
    if concept_lookup is None or target_concept_id not in concept_lookup:
        return [new_line]
    up_to_date_concept = concept_lookup[target_concept_id]
    return [update_row(new_line.copy(), up_to_date_concept, is_new=False)]


def write_usagi_file(
    usagi_file: Path,
    new_mappings: dict[int, NewMap | None],
    overwrite: bool,
    concept_lookup: dict[int, Concept] | None = None,
):
    out_dir = usagi_file.parent
    out_file = out_dir / f"{usagi_file.stem}_{strftime('%Y-%m-%dT%H%M%S')}.csv"

    with usagi_file.open("r") as f_in, out_file.open("w") as f_out:
        reader = csv.DictReader(f_in, delimiter=",")
        out_header = list(reader.fieldnames)
        # Usagi review and STCM files don't contain the mapping type,
        # but it is essential for any MAPS_TO_VALUE relationships.
        if "mappingType" not in out_header:
            out_header.append("mappingType")
        writer = csv.DictWriter(f_out, delimiter=",", fieldnames=out_header)
        writer.writeheader()
        for line in reader:
            new_lines = get_new_lines(line, new_mappings, concept_lookup)
            for new_line in new_lines:
                writer.writerow(new_line)
    if overwrite:
        out_file = out_file.replace(usagi_file)
    logger.info(f"Updated Usagi file available at: {out_file}")


def write_mapping_paths(usagi_file: Path, new_maps: list[NewMap]) -> None:
    out_dir = usagi_file.parent
    out_file = out_dir / f"{usagi_file.stem}_map_path_{strftime('%Y-%m-%dT%H%M%S')}.yml"
    map_paths_as_dict = {}
    for nm in new_maps:
        map_paths_as_dict.update(nm.to_map_path_data())
    write_yaml_file(map_paths_as_dict, out_file)


def write_yaml_file(contents: dict | list, out_path: Path) -> None:
    """Write dict to a YAML file."""
    with out_path.open("w", encoding="utf8") as out:
        yaml.dump(contents, out)



import logging
from collections.abc import Sequence

from omop_cdm.regular.cdm54.tables import Concept

from .relationship import NewMap

logger = logging.getLogger(__name__)


def log_missing_in_db(found_concepts: Sequence[Concept], all_concept_ids: set[int]) -> None:
    concept_ids_in_db = {c.concept_id for c in found_concepts}
    missing_in_db = {c_id for c_id in all_concept_ids if c_id not in concept_ids_in_db}
    if missing_in_db:
        logger.warning(
            f"😕 {len(missing_in_db)} concepts could not be found in the "
            f"database, these will be ignored: {missing_in_db}\n"
        )
    else:
        logger.info("All target concepts found in database 🥳")


def log_remapped_concepts(new_mappings: dict[int, NewMap | None]) -> None:
    n_mapped = len([v for v in new_mappings.values() if v is not None])
    if n_mapped > 0:
        logger.info(f"{n_mapped}/{len(new_mappings)} could be remapped ✨✨")
    else:
        logger.info("No standard concepts could be found ☹️")
    unmapped = sorted([c_id for c_id, v in new_mappings.items() if v is None])
    if unmapped:
        logger.info(f"The following concept_ids could not be remapped: {unmapped}")



import logging

import pytest

from kotobuki.mapping_updater.log import log_remapped_concepts


def test_unmapped_concepts_log_is_sorted(caplog: pytest.LogCaptureFixture):
    new_mappings = {0: None, 99: None, 42: None, 5: None, 100: None}
    with caplog.at_level(logging.INFO):
        log_remapped_concepts(new_mappings)
    assert "The following concept_ids could not be remapped: [0, 5, 42, 99, 100]" in caplog.text



import logging
import sys
from importlib.metadata import version
from pathlib import Path

from omop_cdm.constants import VOCAB_SCHEMA
from sqlalchemy import Engine
from sqlalchemy.orm import Session

from .db import (
    NewMap,
    find_new_mapping,
    query_concepts,
)
from .io import (
    get_target_concepts,
    write_mapping_paths,
    write_usagi_file,
)
from .log import (
    log_missing_in_db,
    log_remapped_concepts,
)

logger = logging.getLogger(__name__)


def update_usagi_file(
    engine: Engine,
    vocab_schema: str,
    usagi_file: Path,
    allow_homonyms: bool = False,
    ignore_case: bool = False,
    write_map_paths: bool = False,
    inspect_only: bool = False,
    overwrite: bool = False,
    update_all: bool = False,
):
    """
    Parse an Usagi exported file to update non-standard concepts.

    :param engine: SQLAlchemy engine to connect with.
    :param vocab_schema: Schema containing the vocabulary tables.
    :param usagi_file: Usagi exported file (save/review/STCM).
    :param allow_homonyms: If no standard concept can be found through
        concept relationships, try to find it through concepts with the
        same concept name. It is strongly recommended to index the
        concept_name column when using this.
    :param write_map_paths: Write an additional file that shows for each
        remapped concept, through which mapping relationships the new
        target concept was determined.
    :param inspect_only: Instead of writing updated mappings to a file,
        only show results in the console.
    :param overwrite: Overwrite the existing Usagi file (otherwise a new
        file is written).
    :param update_all: Also update fields of concepts that don't get a
        new mapping, but do have outdated properties (e.g. changed
        domain_id).
    :return: None
    """
    logging.basicConfig(stream=sys.stdout, format="%(message)s", level=logging.INFO)
    logger.info(f"Running kotobuki v{version('kotobuki')}")

    if inspect_only and overwrite:
        raise ValueError("inspect_only and overwrite cannot both be True.")

    logging.info(f"Parsing Usagi save file {usagi_file.name}")
    concept_ids = get_target_concepts(usagi_file)
    logging.info(f"{len(concept_ids)} distinct target concepts found in file")
    if not concept_ids:
        return

    engine = engine.execution_options(schema_translate_map={VOCAB_SCHEMA: vocab_schema})

    with Session(engine) as session, session.begin():
        concepts = query_concepts(concept_ids, session)
        log_missing_in_db(found_concepts=concepts, all_concept_ids=concept_ids)

        non_standard = {c for c in concepts if c.standard_concept != "S"}
        if not non_standard:
            logger.info("All target concepts are already standard 😍")
            if not update_all:
                return
        else:
            logger.info(f"{len(non_standard)} target concepts are non-standard")

        new_mappings: dict[int, NewMap | None] = {c.concept_id: None for c in non_standard}

        logger.info("Querying database for standard concepts...")
        for concept in non_standard:
            new_map = find_new_mapping(concept, allow_homonyms, ignore_case, session)
            new_mappings[concept.concept_id] = new_map
        log_remapped_concepts(new_mappings)

        if write_map_paths:
            logger.info("Writing mapping paths file")
            new_maps = [nm for nm in new_mappings.values() if nm is not None]
            write_mapping_paths(usagi_file, new_maps)

        if inspect_only:
            return

        logger.info("Writing updated Usagi file")
        if update_all:
            concept_lookup = {c.concept_id: c for c in concepts}
            write_usagi_file(usagi_file, new_mappings, overwrite, concept_lookup)
        else:
            write_usagi_file(usagi_file, new_mappings, overwrite)



from dataclasses import dataclass, field
from enum import Enum

from omop_cdm.regular.cdm54 import Concept


class Relationship(Enum):
    MAPS_TO = "Maps to"
    REPLACED = "Concept replaced by"
    POSS_EQUIVALENT = "Concept poss_eq to"
    SAME_AS = "Concept same_as to"
    MAPS_TO_VALUE = "Maps to value"
    HOMONYM = "Homonym"  # Same concept name (no direct vocabulary link)

    @staticmethod
    def db_relationships() -> set[str]:
        return {
            Relationship.MAPS_TO.value,
            Relationship.REPLACED.value,
            Relationship.POSS_EQUIVALENT.value,
            Relationship.SAME_AS.value,
            Relationship.MAPS_TO_VALUE.value,
        }


# Reverse lookup: enum string to enum
VAL_TO_RELATIONSHIP = {e.value: e for e in Relationship}


@dataclass
class MapLink:
    """Single link in a mapping path.

    :param concept: Concept mapped to.
    :param via: Relationship type that led to the concept, if any.
    """

    concept: Concept
    via: Relationship | None = None

    def __str__(self) -> str:
        s = f"{self.concept.concept_id} {self.concept.concept_name}"
        if self.via is not None:
            s = f"({self.via.value}) {s}"
        return s


@dataclass
class NewMap:
    concepts: list[Concept]
    value_as_concept: list[Concept] = field(default_factory=list)
    map_path: list[MapLink] = field(default_factory=list)

    def __repr__(self):
        s = f"Maps to: {','.join(str(c.concept_id) for c in self.concepts)}"
        if self.value_as_concept:
            s = f"{s}\nMaps to value: {','.join(str(c.concept_id) for c in self.value_as_concept)}"
        return s

    def to_map_path_data(self) -> dict[str, dict[str, list[str]]]:
        """Convert to dict suitable for writing to mapping path file."""
        if not self.map_path:
            return {}
        # we only want to show the intermediate steps in the mapping
        # path, so we ignore the source concept
        map_path = [str(map_link) for map_link in self.map_path[1:]]
        map_properties = {
            "map_path": map_path,
            "maps_to": [_concept_to_str(c) for c in self.concepts],
            "maps_to_value": [_concept_to_str(c) for c in self.value_as_concept],
        }
        # Remove keys that point to empty lists
        map_properties = {k: v for k, v in map_properties.items() if v}

        source_concept = self.map_path[0].concept
        return {_concept_to_str(source_concept): map_properties}


def _concept_to_str(c: Concept) -> str:
    return f"{c.concept_id} {c.concept_name} ({c.vocabulary_id} - {c.domain_id})"



import csv
import filecmp
from pathlib import Path

import pytest
from sqlalchemy import Engine

from kotobuki import update_usagi_file
from tests.python.mapping_updater.conftest import (
    MAP_TO_0_USAGI_FILE,
    USAGI_STCM_FILE,
    write_tmp_usagi_file,
)

pytestmark = pytest.mark.usefixtures("create_vocab_tables")


def test_inspect_only_does_not_change_file(tmp_path: Path, pg_db_engine: Engine):
    tmp_usagi_file = write_tmp_usagi_file(tmp_path, USAGI_STCM_FILE)
    assert len(list(tmp_path.glob("*"))) == 1
    update_usagi_file(pg_db_engine, "vocab", tmp_usagi_file, inspect_only=True)
    assert len(list(tmp_path.glob("*"))) == 1
    assert filecmp.cmp(USAGI_STCM_FILE, tmp_usagi_file)


def test_overwrite(tmp_path: Path, pg_db_engine: Engine):
    tmp_usagi_file = write_tmp_usagi_file(tmp_path, USAGI_STCM_FILE)
    update_usagi_file(pg_db_engine, "vocab", tmp_usagi_file, overwrite=True)
    with tmp_usagi_file.open("r", encoding="utf8") as f:
        reader = csv.DictReader(f)
        mappings = [(line["source_code"], line["target_concept_id"]) for line in reader]
    assert mappings == [
        ("X1", "3"),
        ("X2", "3"),
        ("X3", "3"),
        ("X4", "4"),
        ("X5", "5"),
        ("X6", "6"),
        ("X7", "8"),
        ("X7", "9"),
        ("X8", "8"),
        ("X9", "9"),
        ("X10", "11"),
        ("X10", "12"),
        ("X11", "11"),
        ("X12", "12"),
        ("X13", "14"),
        ("X13", "15"),
        ("X13", "16"),
        ("X14", "14"),
        ("X15", "15"),
        ("X16", "16"),
    ]


def test_homonyms(tmp_path: Path, pg_db_engine: Engine):
    tmp_usagi_file = write_tmp_usagi_file(tmp_path, USAGI_STCM_FILE)
    update_usagi_file(pg_db_engine, "vocab", tmp_usagi_file, overwrite=True, allow_homonyms=True)
    with tmp_usagi_file.open("r", encoding="utf8") as f:
        reader = csv.DictReader(f)
        mappings = [(line["source_code"], line["target_concept_id"]) for line in reader]
    assert mappings == [
        ("X1", "3"),
        ("X2", "3"),
        ("X3", "3"),
        ("X4", "4"),
        ("X5", "6"),  # Now mapped to 6 via a homonym
        ("X6", "6"),
        ("X7", "8"),
        ("X7", "9"),
        ("X8", "8"),
        ("X9", "9"),
        ("X10", "11"),
        ("X10", "12"),
        ("X11", "11"),
        ("X12", "12"),
        ("X13", "14"),
        ("X13", "15"),
        ("X13", "16"),
        ("X14", "14"),
        ("X15", "15"),
        ("X16", "16"),
    ]


def test_domain_id_only_updated_when_mapped_to_non_zero(tmp_path: Path, pg_db_engine: Engine):
    tmp_usagi_save_file = write_tmp_usagi_file(tmp_path, MAP_TO_0_USAGI_FILE)

    def get_mappings_from_file(usagi_file: Path) -> list[tuple[str, str]]:
        with usagi_file.open("r", encoding="utf8") as f:
            reader = csv.DictReader(f)
            return [(line["domainId"], line["conceptId"]) for line in reader]

    expected_original = [
        ("", "0"),
        ("BLA", "0"),
        ("", "0"),
        ("Observation", "17"),
        ("Drug", "18"),
    ]
    assert get_mappings_from_file(MAP_TO_0_USAGI_FILE) == expected_original

    # Without update_all, concept properties of existing mappings are not updated
    update_usagi_file(pg_db_engine, "vocab", tmp_usagi_save_file, overwrite=True, update_all=False)
    assert get_mappings_from_file(tmp_usagi_save_file) == expected_original

    # Codes that map to 0 will not be updated, but codes of which the
    # domain_id is outdated, will get an update
    update_usagi_file(pg_db_engine, "vocab", tmp_usagi_save_file, overwrite=True, update_all=True)
    assert get_mappings_from_file(tmp_usagi_save_file) == [
        ("", "0"),
        ("BLA", "0"),
        ("", "0"),
        ("Condition", "17"),
        ("Condition", "18"),
    ]



import nox  # type: ignore

PYPROJECT = nox.project.load_toml("pyproject.toml")
PYTHON_VERSIONS = nox.project.python_versions(PYPROJECT)

nox.options.default_venv_backend = "uv"

nox.options.sessions = [
    "tests",
    "lint",
]


@nox.session(python=PYTHON_VERSIONS)
def tests(s: nox.Session):
    s.install(".")
    s.run("uv", "pip", "install", "--group", "test")
    s.run("pytest", "--cov-report", "term-missing", "--cov=kotobuki")


@nox.session(python=False, name="format")
def format_all(s: nox.Session):
    """Format codebase with ruff."""
    s.run("uv", "run", "ruff", "format")
    # format imports according to isort via ruff check
    s.run("uv", "run", "ruff", "check", "--select", "I", "--fix")


@nox.session(python=False)
def lint(s: nox.Session):
    """Run ruff linter."""
    s.run("uv", "run", "ruff", "check")
    # Check if any code requires formatting via ruff format
    s.run("uv", "run", "ruff", "format", "--diff")



from collections.abc import Sequence

from omop_cdm.regular.cdm54 import Concept, ConceptRelationship
from sqlalchemy import and_, func, select
from sqlalchemy.orm import Session

from .relationship import (
    VAL_TO_RELATIONSHIP,
    MapLink,
    NewMap,
    Relationship,
)


def query_concepts(concept_ids: set[int], session: Session) -> Sequence[Concept]:
    """Get Concept ORM objects for all given concept_ids."""
    return session.scalars(select(Concept).filter(Concept.concept_id.in_(concept_ids))).all()


def get_mappings(concept_id: int, session: Session) -> Sequence[ConceptRelationship]:
    """Get relationship mappings for a given concept_id."""
    return session.scalars(
        select(ConceptRelationship).filter(
            and_(
                ConceptRelationship.concept_id_1 == concept_id,
                ConceptRelationship.relationship_id.in_(Relationship.db_relationships()),
            )
        )
    ).all()


def find_maps_to_value_relationship(
    mappings: Sequence[ConceptRelationship],
) -> list[Concept]:
    return [m.concept_2 for m in mappings if m.relationship_id == Relationship.MAPS_TO_VALUE.value]


def find_standard_concepts(
    concept_id: int, session: Session, path: list[MapLink]
) -> NewMap | None:
    """Search recursively for a standard concept."""
    mappings = get_mappings(concept_id, session)
    maps_to_mappings = [m for m in mappings if m.relationship_id == Relationship.MAPS_TO.value]

    # If there is one (or more) Maps to relationship, we are done
    if maps_to_mappings:
        target_concepts = [c.concept_2 for c in maps_to_mappings]
        result = NewMap(concepts=target_concepts, map_path=path)
        result.value_as_concept = find_maps_to_value_relationship(mappings)
        return result

    for mapping in mappings:
        if mapping.relationship_id in {
            Relationship.REPLACED.value,
            Relationship.POSS_EQUIVALENT.value,
            Relationship.SAME_AS.value,
        }:
            new_path = path.copy()
            via = VAL_TO_RELATIONSHIP[mapping.relationship_id]
            new_path.append(MapLink(mapping.concept_2, via))
            return find_standard_concepts(mapping.concept_id_2, session, new_path)
    return None


def find_all_homonyms(
    concept: Concept, case_insensitive: bool, session: Session
) -> Sequence[Concept]:
    """Get Concept ORM objects that match the concept_name."""
    if case_insensitive:
        homonyms = session.scalars(
            select(Concept).where(
                and_(
                    func.lower(Concept.concept_name) == concept.concept_name.lower(),
                    Concept.concept_id != concept.concept_id,
                )
            )
        ).all()
    else:
        homonyms = session.scalars(
            select(Concept).where(
                and_(
                    Concept.concept_name == concept.concept_name,
                    Concept.concept_id != concept.concept_id,
                )
            )
        ).all()
    return homonyms


def find_suitable_homonym(
    homonyms: Sequence[Concept], session: Session, concept: Concept
) -> NewMap | None:
    """
    Return a new mapping for a homonym that maps to a standard concept.

    If multiple homonyms map to a standard concept, those that are from
    the same domain as the original concept are prioritized. If there
    are no homonyms, or none map to a standard concept, return None.
    """
    start_path = [MapLink(concept)]
    mappings: list[NewMap] = []
    for h in homonyms:
        path = start_path.copy()
        path.append(MapLink(h, Relationship.HOMONYM))
        new_map = find_standard_concepts(h.concept_id, session, path)
        if new_map is not None:
            mappings.append(new_map)
    same_domain_mappings = [
        nm for nm in mappings if any(c.domain_id == concept.domain_id for c in nm.concepts)
    ]
    if same_domain_mappings:
        return same_domain_mappings[0]
    if mappings:
        return mappings[0]
    return None


def find_new_mapping(
    concept: Concept,
    search_homonyms: bool,
    ignore_case: bool,
    session: Session,
) -> NewMap | None:
    # Try to find standard concepts via concept relationships
    new_map = find_standard_concepts(concept.concept_id, session, [MapLink(concept)])
    # Alternatively via concepts with an identical name
    if search_homonyms and new_map is None:
        homonyms = find_all_homonyms(concept, ignore_case, session)
        new_map = find_suitable_homonym(homonyms, session, concept)
    return new_map



import pytest
from sqlalchemy import Engine

from tests.python.mapping_updater.test_usagi_mappings import get_new_map

pytestmark = pytest.mark.usefixtures("create_vocab_tables")


def test_homonym_mapping_active(pg_db_engine: Engine):
    """Concept 5 can only map to 6 via a homonym."""
    result = get_new_map(concept_id=5, engine=pg_db_engine, homonyms=True)
    # With activated homonym search, it should map to 6
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 6
    assert not result.value_as_concept


def test_homonym_with_same_domain_gets_priority(pg_db_engine: Engine):
    """Of multiple standard homonyms, only concept 24 is from the same domain."""
    result = get_new_map(concept_id=21, engine=pg_db_engine, homonyms=True, ignore_case=True)
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 24
    assert not result.value_as_concept


def test_find_standard_concept_via_case_insensitive_homonym(pg_db_engine: Engine):
    """
    The concept QzErTy is non-standard without a relationship to a standard concept,
    however, there is the standard homonym "qzerty". Test if this homonym is picked up correctly.
    """
    result = get_new_map(
        concept_id=19,
        engine=pg_db_engine,
        homonyms=True,
        ignore_case=True,
    )
    assert len(result.concepts) == 1
    assert result.concepts[0].concept_id == 20
    assert not result.value_as_concept


def test_ignore_case_false(pg_db_engine: Engine):
    """
    Test if "qzerty" is not picked up from "QzErTy" when ignore_case is False.
    """
    result = get_new_map(
        concept_id=19,
        engine=pg_db_engine,
        homonyms=True,
        ignore_case=False,
    )
    assert result is None
